@@ -18,33 +18,36 @@
 import random
 import time
 
-import six
+from google.api_core import gapic_v1  # type: ignore
+from google.api_core import retry as retries  # type: ignore
 
-from google.api_core import exceptions
+from google.cloud.firestore_v1.base_transaction import (
+    _BaseTransactional,
+    BaseTransaction,
+    MAX_ATTEMPTS,
+    _CANT_BEGIN,
+    _CANT_ROLLBACK,
+    _CANT_COMMIT,
+    _WRITE_READ_ONLY,
+    _INITIAL_SLEEP,
+    _MAX_SLEEP,
+    _MULTIPLIER,
+    _EXCEED_ATTEMPTS_TEMPLATE,
+)
+
+from google.api_core import exceptions  # type: ignore
 from google.cloud.firestore_v1 import batch
-from google.cloud.firestore_v1 import types
 from google.cloud.firestore_v1.document import DocumentReference
+from google.cloud.firestore_v1 import _helpers
 from google.cloud.firestore_v1.query import Query
 
-
-MAX_ATTEMPTS = 5
-"""int: Default number of transaction attempts (with retries)."""
-_CANT_BEGIN = "The transaction has already begun. Current transaction ID: {!r}."
-_MISSING_ID_TEMPLATE = "The transaction has no transaction ID, so it cannot be {}."
-_CANT_ROLLBACK = _MISSING_ID_TEMPLATE.format("rolled back")
-_CANT_COMMIT = _MISSING_ID_TEMPLATE.format("committed")
-_WRITE_READ_ONLY = "Cannot perform write operation in read-only transaction."
-_INITIAL_SLEEP = 1.0
-"""float: Initial "max" for sleep interval. To be used in :func:`_sleep`."""
-_MAX_SLEEP = 30.0
-"""float: Eventual "max" sleep time. To be used in :func:`_sleep`."""
-_MULTIPLIER = 2.0
-"""float: Multiplier for exponential backoff. To be used in :func:`_sleep`."""
-_EXCEED_ATTEMPTS_TEMPLATE = "Failed to commit transaction in {:d} attempts."
-_CANT_RETRY_READ_ONLY = "Only read-write transactions can be retried."
+# Types needed only for Type Hints
+from google.cloud.firestore_v1.base_document import DocumentSnapshot
+from google.cloud.firestore_v1.types import CommitResponse
+from typing import Any, Callable, Generator, Optional
 
 
-class Transaction(batch.WriteBatch):
+class Transaction(batch.WriteBatch, BaseTransaction):
     """Accumulate read-and-write operations to be sent in a transaction.
 
     Args:
@@ -58,18 +61,16 @@ class Transaction(batch.WriteBatch):
             :data:`False`.
     """
 
-    def __init__(self, client, max_attempts=MAX_ATTEMPTS, read_only=False):
+    def __init__(self, client, max_attempts=MAX_ATTEMPTS, read_only=False) -> None:
         super(Transaction, self).__init__(client)
-        self._max_attempts = max_attempts
-        self._read_only = read_only
-        self._id = None
+        BaseTransaction.__init__(self, max_attempts, read_only)
 
-    def _add_write_pbs(self, write_pbs):
+    def _add_write_pbs(self, write_pbs: list) -> None:
         """Add `Write`` protobufs to this transaction.
 
         Args:
             write_pbs (List[google.cloud.proto.firestore.v1.\
-                write_pb2.Write]): A list of write protobufs to be added.
+                write.Write]): A list of write protobufs to be added.
 
         Raises:
             ValueError: If this transaction is read-only.
@@ -79,62 +80,7 @@ class Transaction(batch.WriteBatch):
 
         super(Transaction, self)._add_write_pbs(write_pbs)
 
-    def _options_protobuf(self, retry_id):
-        """Convert the current object to protobuf.
-
-        The ``retry_id`` value is used when retrying a transaction that
-        failed (e.g. due to contention). It is intended to be the "first"
-        transaction that failed (i.e. if multiple retries are needed).
-
-        Args:
-            retry_id (Union[bytes, NoneType]): Transaction ID of a transaction
-                to be retried.
-
-        Returns:
-            Optional[google.cloud.firestore_v1.types.TransactionOptions]:
-            The protobuf ``TransactionOptions`` if ``read_only==True`` or if
-            there is a transaction ID to be retried, else :data:`None`.
-
-        Raises:
-            ValueError: If ``retry_id`` is not :data:`None` but the
-                transaction is read-only.
-        """
-        if retry_id is not None:
-            if self._read_only:
-                raise ValueError(_CANT_RETRY_READ_ONLY)
-
-            return types.TransactionOptions(
-                read_write=types.TransactionOptions.ReadWrite(
-                    retry_transaction=retry_id
-                )
-            )
-        elif self._read_only:
-            return types.TransactionOptions(
-                read_only=types.TransactionOptions.ReadOnly()
-            )
-        else:
-            return None
-
-    @property
-    def in_progress(self):
-        """Determine if this transaction has already begun.
-
-        Returns:
-            bool: Indicates if the transaction has started.
-        """
-        return self._id is not None
-
-    @property
-    def id(self):
-        """Get the current transaction ID.
-
-        Returns:
-            Optional[bytes]: The transaction ID (or :data:`None` if the
-            current transaction is not in progress).
-        """
-        return self._id
-
-    def _begin(self, retry_id=None):
+    def _begin(self, retry_id: bytes = None) -> None:
         """Begin the transaction.
 
         Args:
@@ -149,21 +95,15 @@ class Transaction(batch.WriteBatch):
             raise ValueError(msg)
 
         transaction_response = self._client._firestore_api.begin_transaction(
-            self._client._database_string,
-            options_=self._options_protobuf(retry_id),
+            request={
+                "database": self._client._database_string,
+                "options": self._options_protobuf(retry_id),
+            },
             metadata=self._client._rpc_metadata,
         )
         self._id = transaction_response.transaction
 
-    def _clean_up(self):
-        """Clean up the instance after :meth:`_rollback`` or :meth:`_commit``.
-
-        This intended to occur on success or failure of the associated RPCs.
-        """
-        self._write_pbs = []
-        self._id = None
-
-    def _rollback(self):
+    def _rollback(self) -> None:
         """Roll back the transaction.
 
         Raises:
@@ -175,18 +115,20 @@ class Transaction(batch.WriteBatch):
         try:
             # NOTE: The response is just ``google.protobuf.Empty``.
             self._client._firestore_api.rollback(
-                self._client._database_string,
-                self._id,
+                request={
+                    "database": self._client._database_string,
+                    "transaction": self._id,
+                },
                 metadata=self._client._rpc_metadata,
             )
         finally:
             self._clean_up()
 
-    def _commit(self):
+    def _commit(self) -> list:
         """Transactionally commit the changes accumulated.
 
         Returns:
-            List[:class:`google.cloud.proto.firestore.v1.write_pb2.WriteResult`, ...]:
+            List[:class:`google.cloud.proto.firestore.v1.write.WriteResult`, ...]:
             The write results corresponding to the changes committed, returned
             in the same order as the changes were applied to this transaction.
             A write result contains an ``update_time`` field.
@@ -202,39 +144,60 @@ class Transaction(batch.WriteBatch):
         self._clean_up()
         return list(commit_response.write_results)
 
-    def get_all(self, references):
+    def get_all(
+        self,
+        references: list,
+        retry: retries.Retry = gapic_v1.method.DEFAULT,
+        timeout: float = None,
+    ) -> Generator[DocumentSnapshot, Any, None]:
         """Retrieves multiple documents from Firestore.
 
         Args:
             references (List[.DocumentReference, ...]): Iterable of document
                 references to be retrieved.
+            retry (google.api_core.retry.Retry): Designation of what errors, if any,
+                should be retried.  Defaults to a system-specified policy.
+            timeout (float): The timeout for this request.  Defaults to a
+                system-specified value.
 
         Yields:
             .DocumentSnapshot: The next document snapshot that fulfills the
             query, or :data:`None` if the document does not exist.
         """
-        return self._client.get_all(references, transaction=self)
+        kwargs = _helpers.make_retry_timeout_kwargs(retry, timeout)
+        return self._client.get_all(references, transaction=self, **kwargs)
+
+    def get(
+        self,
+        ref_or_query,
+        retry: retries.Retry = gapic_v1.method.DEFAULT,
+        timeout: float = None,
+    ) -> Generator[DocumentSnapshot, Any, None]:
+        """Retrieve a document or a query result from the database.
 
-    def get(self, ref_or_query):
-        """
-        Retrieve a document or a query result from the database.
         Args:
-            ref_or_query The document references or query object to return.
+            ref_or_query: The document references or query object to return.
+            retry (google.api_core.retry.Retry): Designation of what errors, if any,
+                should be retried.  Defaults to a system-specified policy.
+            timeout (float): The timeout for this request.  Defaults to a
+                system-specified value.
+
         Yields:
             .DocumentSnapshot: The next document snapshot that fulfills the
             query, or :data:`None` if the document does not exist.
         """
+        kwargs = _helpers.make_retry_timeout_kwargs(retry, timeout)
         if isinstance(ref_or_query, DocumentReference):
-            return self._client.get_all([ref_or_query], transaction=self)
+            return self._client.get_all([ref_or_query], transaction=self, **kwargs)
         elif isinstance(ref_or_query, Query):
-            return ref_or_query.stream(transaction=self)
+            return ref_or_query.stream(transaction=self, **kwargs)
         else:
             raise ValueError(
                 'Value for argument "ref_or_query" must be a DocumentReference or a Query.'
             )
 
 
-class _Transactional(object):
+class _Transactional(_BaseTransactional):
     """Provide a callable object to use as a transactional decorater.
 
     This is surfaced via
@@ -245,19 +208,10 @@ class _Transactional(object):
             A callable that should be run (and retried) in a transaction.
     """
 
-    def __init__(self, to_wrap):
-        self.to_wrap = to_wrap
-        self.current_id = None
-        """Optional[bytes]: The current transaction ID."""
-        self.retry_id = None
-        """Optional[bytes]: The ID of the first attempted transaction."""
+    def __init__(self, to_wrap) -> None:
+        super(_Transactional, self).__init__(to_wrap)
 
-    def _reset(self):
-        """Unset the transaction IDs."""
-        self.current_id = None
-        self.retry_id = None
-
-    def _pre_commit(self, transaction, *args, **kwargs):
+    def _pre_commit(self, transaction: Transaction, *args, **kwargs) -> Any:
         """Begin transaction and call the wrapped callable.
 
         If the callable raises an exception, the transaction will be rolled
@@ -295,7 +249,7 @@ class _Transactional(object):
             transaction._rollback()
             raise
 
-    def _maybe_commit(self, transaction):
+    def _maybe_commit(self, transaction: Transaction) -> Optional[bool]:
         """Try to commit the transaction.
 
         If the transaction is read-write and the ``Commit`` fails with the
@@ -323,7 +277,7 @@ class _Transactional(object):
             else:
                 raise
 
-    def __call__(self, transaction, *args, **kwargs):
+    def __call__(self, transaction: Transaction, *args, **kwargs):
         """Execute the wrapped callable within a transaction.
 
         Args:
@@ -344,7 +298,7 @@ class _Transactional(object):
         """
         self._reset()
 
-        for attempt in six.moves.xrange(transaction._max_attempts):
+        for attempt in range(transaction._max_attempts):
             result = self._pre_commit(transaction, *args, **kwargs)
             succeeded = self._maybe_commit(transaction)
             if succeeded:
@@ -361,7 +315,7 @@ class _Transactional(object):
         raise ValueError(msg)
 
 
-def transactional(to_wrap):
+def transactional(to_wrap: Callable) -> _Transactional:
     """Decorate a callable so that it runs in a transaction.
 
     Args:
@@ -376,7 +330,9 @@ def transactional(to_wrap):
     return _Transactional(to_wrap)
 
 
-def _commit_with_retry(client, write_pbs, transaction_id):
+def _commit_with_retry(
+    client, write_pbs: list, transaction_id: bytes
+) -> CommitResponse:
     """Call ``Commit`` on the GAPIC client with retry / sleep.
 
     Retries the ``Commit`` RPC on Unavailable. Usually this RPC-level
@@ -388,7 +344,7 @@ def _commit_with_retry(client, write_pbs, transaction_id):
     Args:
         client (:class:`~google.cloud.firestore_v1.client.Client`):
             A client with GAPIC client and configuration details.
-        write_pbs (List[:class:`google.cloud.proto.firestore.v1.write_pb2.Write`, ...]):
+        write_pbs (List[:class:`google.cloud.proto.firestore.v1.write.Write`, ...]):
             A ``Write`` protobuf instance to be committed.
         transaction_id (bytes):
             ID of an existing transaction that this commit will run in.
@@ -405,9 +361,11 @@ def _commit_with_retry(client, write_pbs, transaction_id):
     while True:
         try:
             return client._firestore_api.commit(
-                client._database_string,
-                write_pbs,
-                transaction=transaction_id,
+                request={
+                    "database": client._database_string,
+                    "writes": write_pbs,
+                    "transaction": transaction_id,
+                },
                 metadata=client._rpc_metadata,
             )
         except exceptions.ServiceUnavailable:
@@ -417,7 +375,9 @@ def _commit_with_retry(client, write_pbs, transaction_id):
         current_sleep = _sleep(current_sleep)
 
 
-def _sleep(current_sleep, max_sleep=_MAX_SLEEP, multiplier=_MULTIPLIER):
+def _sleep(
+    current_sleep: float, max_sleep: float = _MAX_SLEEP, multiplier: float = _MULTIPLIER
+) -> float:
     """Sleep and produce a new sleep time.
 
     .. _Exponential Backoff And Jitter: https://www.awsarchitectureblog.com/\
